package java.security;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import java.io.ByteArrayOutputStream;
import javax.crypto.NoSuchPaddingException;
import java.util.concurrent.ConcurrentHashMap;
import java.util.Arrays;
import sun.misc.SharedSecrets;
import sun.misc.JavaSecuritySignatureAccess;
import java.nio.ByteBuffer;
import java.util.Set;
import java.security.cert.X509Certificate;
import java.security.cert.Certificate;
import java.security.spec.AlgorithmParameterSpec;
import javax.crypto.Cipher;
import java.util.Iterator;
import sun.security.jca.GetInstance;
import java.util.Map;
import sun.security.jca.ServiceId;
import java.util.List;
import sun.security.util.Debug;

public abstract class Signature extends SignatureSpi
{
    private static final Debug debug;
    private static final Debug pdebug;
    private static final boolean skipDebug;
    private String algorithm;
    Provider provider;
    protected static final int UNINITIALIZED = 0;
    protected static final int SIGN = 2;
    protected static final int VERIFY = 3;
    protected int state;
    private static final String RSA_SIGNATURE = "NONEwithRSA";
    private static final String RSA_CIPHER = "RSA/ECB/PKCS1Padding";
    private static final List<ServiceId> rsaIds;
    private static final Map<String, Boolean> signatureInfo;
    
    protected Signature(String algorithm) {
        this.state = 0;
        this.algorithm = algorithm;
    }
    
    public static Signature getInstance(String s) throws NoSuchAlgorithmException {
        List<Provider.Service> list;
        if (s.equalsIgnoreCase("NONEwithRSA")) {
            list = GetInstance.getServices(Signature.rsaIds);
        }
        else {
            list = GetInstance.getServices("Signature", s);
        }
        Iterator<Object> iterator = list.iterator();
        if (!iterator.hasNext()) {
            throw new NoSuchAlgorithmException(s + " Signature not available");
        }
        Provider.Service s2;
        while (true) {
            s2 = (Provider.Service)iterator.next();
            if (isSpi(s2)) {
                break;
            }
            try {
                return getInstance(GetInstance.getInstance(s2, SignatureSpi.class), s);
            }
            catch (NoSuchAlgorithmException ex) {
                if (iterator.hasNext()) {
                    continue;
                }
                throw ex;
            }
        }
        return new Signature.Delegate(s2, iterator, s);
    }
    
    private static Signature getInstance(GetInstance.Instance instance, String algorithm) {
        Signature signature;
        if ((instance.impl instanceof Signature)) {
            signature = (Signature)instance.impl;
            signature.algorithm = algorithm;
        }
        else {
            signature = new Signature.Delegate((SignatureSpi)instance.impl, algorithm);
        }
        signature.provider = instance.provider;
        return signature;
    }
    
    private static boolean isSpi(Provider.Service service) {
        if (service.getType().equals("Cipher")) {
            return true;
        }
        String className = service.getClassName();
        Boolean value = (Boolean)Signature.signatureInfo.get(className);
        if (value == null) {
            try {
                Object instance = service.newInstance(null);
                int b = instance instanceof SignatureSpi && !(instance instanceof Signature);
                if (Signature.debug != null && b == 0) {
                    Signature.debug.println("Not a SignatureSpi " + className);
                    Signature.debug.println("Delayed provider selection may not be available for algorithm " + service.getAlgorithm());
                }
                value = Boolean.valueOf((boolean)b);
                Signature.signatureInfo.put(className, value);
            }
            catch (Exception ex) {
                return false;
            }
        }
        return (boolean)value;
    }
    
    public static Signature getInstance(String algorithm, String s) throws NoSuchAlgorithmException, NoSuchProviderException {
        if (algorithm.equalsIgnoreCase("NONEwithRSA")) {
            if (s == null || s.length() == 0) {
                throw new IllegalArgumentException("missing provider");
            }
            Provider provider = Security.getProvider(s);
            if (provider == null) {
                throw new NoSuchProviderException("no such provider: " + s);
            }
            return getInstanceRSA(provider);
        }
        else {
            return getInstance(GetInstance.getInstance("Signature", SignatureSpi.class, algorithm, s), algorithm);
        }
    }
    
    public static Signature getInstance(String algorithm, Provider provider) throws NoSuchAlgorithmException {
        if (algorithm.equalsIgnoreCase("NONEwithRSA")) {
            if (provider == null) {
                throw new IllegalArgumentException("missing provider");
            }
            return getInstanceRSA(provider);
        }
        else {
            return getInstance(GetInstance.getInstance("Signature", SignatureSpi.class, algorithm, provider), algorithm);
        }
    }
    
    private static Signature getInstanceRSA(Provider provider) throws NoSuchAlgorithmException {
        Provider.Service service = provider.getService("Signature", "NONEwithRSA");
        if (service != null) {
            return getInstance(GetInstance.getInstance(service, SignatureSpi.class), "NONEwithRSA");
        }
        try {
            return new Signature.Delegate(new Signature.CipherAdapter(Cipher.getInstance("RSA/ECB/PKCS1Padding", provider)), "NONEwithRSA");
        }
        catch (GeneralSecurityException ex) {
            throw new NoSuchAlgorithmException("no such algorithm: NONEwithRSA for provider " + provider.getName(), ex);
        }
    }
    
    public final Provider getProvider() {
        this.chooseFirstProvider();
        return this.provider;
    }
    
    private String getProviderName() {
        return (this.provider == null) ? "(no provider)" : this.provider.getName();
    }
    
    void chooseFirstProvider() {
    }
    
    public final void initVerify(PublicKey publicKey) throws InvalidKeyException {
        this.engineInitVerify(publicKey);
        this.state = 3;
        if (!Signature.skipDebug && Signature.pdebug != null) {
            Signature.pdebug.println("Signature." + this.algorithm + " verification algorithm from: " + this.getProviderName());
        }
    }
    
    final void initVerify(PublicKey publicKey, AlgorithmParameterSpec algorithmParameterSpec) throws InvalidKeyException, InvalidAlgorithmParameterException {
        this.engineInitVerify(publicKey, algorithmParameterSpec);
        this.state = 3;
        if (!Signature.skipDebug && Signature.pdebug != null) {
            Signature.pdebug.println("Signature." + this.algorithm + " verification algorithm from: " + this.getProviderName());
        }
    }
    
    private static PublicKey getPublicKeyFromCert(Certificate certificate) throws InvalidKeyException {
        if ((certificate instanceof X509Certificate)) {
            X509Certificate x509Certificate = (X509Certificate)certificate;
            Set<String> criticalExtensionOIDs = x509Certificate.getCriticalExtensionOIDs();
            if (criticalExtensionOIDs != null && !criticalExtensionOIDs.isEmpty() && criticalExtensionOIDs.contains("2.5.29.15")) {
                boolean[] keyUsage = x509Certificate.getKeyUsage();
                if (keyUsage != null && !keyUsage[0]) {
                    throw new InvalidKeyException("Wrong key usage");
                }
            }
        }
        return certificate.getPublicKey();
    }
    
    public final void initVerify(Certificate certificate) throws InvalidKeyException {
        this.engineInitVerify(getPublicKeyFromCert(certificate));
        this.state = 3;
        if (!Signature.skipDebug && Signature.pdebug != null) {
            Signature.pdebug.println("Signature." + this.algorithm + " verification algorithm from: " + this.getProviderName());
        }
    }
    
    final void initVerify(Certificate certificate, AlgorithmParameterSpec algorithmParameterSpec) throws InvalidKeyException, InvalidAlgorithmParameterException {
        this.engineInitVerify(getPublicKeyFromCert(certificate), algorithmParameterSpec);
        this.state = 3;
        if (!Signature.skipDebug && Signature.pdebug != null) {
            Signature.pdebug.println("Signature." + this.algorithm + " verification algorithm from: " + this.getProviderName());
        }
    }
    
    public final void initSign(PrivateKey privateKey) throws InvalidKeyException {
        this.engineInitSign(privateKey);
        this.state = 2;
        if (!Signature.skipDebug && Signature.pdebug != null) {
            Signature.pdebug.println("Signature." + this.algorithm + " signing algorithm from: " + this.getProviderName());
        }
    }
    
    public final void initSign(PrivateKey privateKey, SecureRandom secureRandom) throws InvalidKeyException {
        this.engineInitSign(privateKey, secureRandom);
        this.state = 2;
        if (!Signature.skipDebug && Signature.pdebug != null) {
            Signature.pdebug.println("Signature." + this.algorithm + " signing algorithm from: " + this.getProviderName());
        }
    }
    
    final void initSign(PrivateKey privateKey, AlgorithmParameterSpec algorithmParameterSpec, SecureRandom secureRandom) throws InvalidKeyException, InvalidAlgorithmParameterException {
        this.engineInitSign(privateKey, algorithmParameterSpec, secureRandom);
        this.state = 2;
        if (!Signature.skipDebug && Signature.pdebug != null) {
            Signature.pdebug.println("Signature." + this.algorithm + " signing algorithm from: " + this.getProviderName());
        }
    }
    
    public final byte[] sign() throws SignatureException {
        if (this.state == 2) {
            return this.engineSign();
        }
        throw new SignatureException("object not initialized for signing");
    }
    
    public final int sign(byte[] array, int n, int n2) throws SignatureException {
        if (array == null) {
            throw new IllegalArgumentException("No output buffer given");
        }
        if (n < 0 || n2 < 0) {
            throw new IllegalArgumentException("offset or len is less than 0");
        }
        if (array.length - n < n2) {
            throw new IllegalArgumentException("Output buffer too small for specified offset and length");
        }
        if (this.state != 2) {
            throw new SignatureException("object not initialized for signing");
        }
        return this.engineSign(array, n, n2);
    }
    
    public final boolean verify(byte[] array) throws SignatureException {
        if (this.state == 3) {
            return this.engineVerify(array);
        }
        throw new SignatureException("object not initialized for verification");
    }
    
    public final boolean verify(byte[] array, int n, int n2) throws SignatureException {
        if (this.state == 3) {
            if (array == null) {
                throw new IllegalArgumentException("signature is null");
            }
            if (n < 0 || n2 < 0) {
                throw new IllegalArgumentException("offset or length is less than 0");
            }
            if (array.length - n < n2) {
                throw new IllegalArgumentException("signature too small for specified offset and length");
            }
            return this.engineVerify(array, n, n2);
        }
        else {
            throw new SignatureException("object not initialized for verification");
        }
    }
    
    public final void update(byte b) throws SignatureException {
        if (this.state == 3 || this.state == 2) {
            this.engineUpdate(b);
            return;
        }
        throw new SignatureException("object not initialized for signature or verification");
    }
    
    public final void update(byte[] array) throws SignatureException {
        this.update(array, 0, array.length);
    }
    
    public final void update(byte[] array, int n, int n2) throws SignatureException {
        if (this.state == 2 || this.state == 3) {
            if (array == null) {
                throw new IllegalArgumentException("data is null");
            }
            if (n < 0 || n2 < 0) {
                throw new IllegalArgumentException("off or len is less than 0");
            }
            if (array.length - n < n2) {
                throw new IllegalArgumentException("data too small for specified offset and length");
            }
            this.engineUpdate(array, n, n2);
            return;
        }
        else {
            throw new SignatureException("object not initialized for signature or verification");
        }
    }
    
    public final void update(ByteBuffer byteBuffer) throws SignatureException {
        if (this.state != 2 && this.state != 3) {
            throw new SignatureException("object not initialized for signature or verification");
        }
        if (byteBuffer == null) {
            throw new NullPointerException();
        }
        this.engineUpdate(byteBuffer);
        return;
    }
    
    public final String getAlgorithm() {
        return this.algorithm;
    }
    
    @Override
    public String toString() {
        String str = "";
        switch (this.state) {
            case 0: {
                str = "<not initialized>";
                break;
            }
            case 3: {
                str = "<initialized for verifying>";
                break;
            }
            case 2: {
                str = "<initialized for signing>";
                break;
            }
        }
        return "Signature object: " + this.getAlgorithm() + str;
    }
    
    @Deprecated
    public final void setParameter(String s, Object o) throws InvalidParameterException {
        this.engineSetParameter(s, o);
    }
    
    public final void setParameter(AlgorithmParameterSpec algorithmParameterSpec) throws InvalidAlgorithmParameterException {
        this.engineSetParameter(algorithmParameterSpec);
    }
    
    public final AlgorithmParameters getParameters() {
        return this.engineGetParameters();
    }
    
    @Deprecated
    public final Object getParameter(String s) throws InvalidParameterException {
        return this.engineGetParameter(s);
    }
    
    @Override
    public Object clone() throws CloneNotSupportedException {
        if ((this instanceof Cloneable)) {
            return super.clone();
        }
        throw new CloneNotSupportedException();
    }
    
    static boolean access$200(Provider.Service service) {
        return isSpi(service);
    }
    
    static {
        SharedSecrets.setJavaSecuritySignatureAccess(new JavaSecuritySignatureAccess() {
            public void initVerify(Signature signature, PublicKey publicKey, AlgorithmParameterSpec algorithmParameterSpec) throws InvalidKeyException, InvalidAlgorithmParameterException {
                signature.initVerify(publicKey, algorithmParameterSpec);
            }
            
            public void initVerify(Signature signature, Certificate certificate, AlgorithmParameterSpec algorithmParameterSpec) throws InvalidKeyException, InvalidAlgorithmParameterException {
                signature.initVerify(certificate, algorithmParameterSpec);
            }
            
            public void initSign(Signature signature, PrivateKey privateKey, AlgorithmParameterSpec algorithmParameterSpec, SecureRandom secureRandom) throws InvalidKeyException, InvalidAlgorithmParameterException {
                signature.initSign(privateKey, algorithmParameterSpec, secureRandom);
            }
        });
        debug = Debug.getInstance("jca", "Signature");
        pdebug = Debug.getInstance("provider", "Provider");
        skipDebug = (Debug.isOn("engine=") && !Debug.isOn("signature"));
        rsaIds = Arrays.asList(new ServiceId[] { new ServiceId("Signature", "NONEwithRSA"), new ServiceId("Cipher", "RSA/ECB/PKCS1Padding"), new ServiceId("Cipher", "RSA/ECB"), new ServiceId("Cipher", "RSA//PKCS1Padding"), new ServiceId("Cipher", "RSA") });
        signatureInfo = new ConcurrentHashMap();
        Boolean true = Boolean.TRUE;
        Signature.signatureInfo.put("sun.security.provider.DSA$RawDSA", true);
        Signature.signatureInfo.put("sun.security.provider.DSA$SHA1withDSA", true);
        Signature.signatureInfo.put("sun.security.rsa.RSASignature$MD2withRSA", true);
        Signature.signatureInfo.put("sun.security.rsa.RSASignature$MD5withRSA", true);
        Signature.signatureInfo.put("sun.security.rsa.RSASignature$SHA1withRSA", true);
        Signature.signatureInfo.put("sun.security.rsa.RSASignature$SHA256withRSA", true);
        Signature.signatureInfo.put("sun.security.rsa.RSASignature$SHA384withRSA", true);
        Signature.signatureInfo.put("sun.security.rsa.RSASignature$SHA512withRSA", true);
        Signature.signatureInfo.put("sun.security.rsa.RSAPSSSignature", true);
        Signature.signatureInfo.put("com.sun.net.ssl.internal.ssl.RSASignature", true);
        Signature.signatureInfo.put("sun.security.pkcs11.P11Signature", true);
    }
    
    private static class Delegate extends Signature
    {
        private SignatureSpi sigSpi;
        private final Object lock;
        private Provider.Service firstService;
        private Iterator<Provider.Service> serviceIterator;
        private static int warnCount;
        private static final int I_PUB = 1;
        private static final int I_PRIV = 2;
        private static final int I_PRIV_SR = 3;
        private static final int I_PUB_PARAM = 4;
        private static final int I_PRIV_PARAM_SR = 5;
        private static final int S_PARAM = 6;
        
        Delegate(SignatureSpi sigSpi, String s) {
            super(s);
            this.sigSpi = sigSpi;
            this.lock = null;
        }
        
        Delegate(Provider.Service firstService, Iterator<Provider.Service> serviceIterator, String s) {
            super(s);
            this.firstService = firstService;
            this.serviceIterator = serviceIterator;
            this.lock = new Object();
        }
        
        @Override
        public Object clone() throws CloneNotSupportedException {
            this.chooseFirstProvider();
            if ((this.sigSpi instanceof Cloneable)) {
                Signature.Delegate delegate = new Signature.Delegate((SignatureSpi)this.sigSpi.clone(), this.algorithm);
                delegate.provider = super.provider;
                return delegate;
            }
            throw new CloneNotSupportedException();
        }
        
        private static SignatureSpi newInstance(Provider.Service service) throws NoSuchAlgorithmException {
            if (service.getType().equals("Cipher")) {
                try {
                    return new Signature.CipherAdapter(Cipher.getInstance("RSA/ECB/PKCS1Padding", service.getProvider()));
                }
                catch (NoSuchPaddingException ex) {
                    throw new NoSuchAlgorithmException(ex);
                }
            }
            else {
                Object instance = service.newInstance(null);
                if (!(instance instanceof SignatureSpi)) {
                    throw new NoSuchAlgorithmException("Not a SignatureSpi: " + instance.getClass().getName());
                }
                return (SignatureSpi)instance;
            }
        }
        
        @Override
        void chooseFirstProvider() {
            if (this.sigSpi != null) {
                return;
            }
            synchronized (this.lock) {
                if (this.sigSpi != null) {
                    return;
                }
                if (Signature.debug != null) {
                    int n = --Signature.Delegate.warnCount;
                    if (n >= 0) {
                        Signature.debug.println("Signature.init() not first method called, disabling delayed provider selection");
                        if (n == 0) {
                            Signature.debug.println("Further warnings of this type will be suppressed");
                        }
                        new Exception("Debug call trace").printStackTrace();
                    }
                }
                NoSuchAlgorithmException cause = null;
                while (this.firstService != null || this.serviceIterator.hasNext()) {
                    Provider.Service firstService;
                    if (this.firstService != null) {
                        firstService = this.firstService;
                        this.firstService = null;
                    }
                    else {
                        firstService = (Provider.Service)this.serviceIterator.next();
                    }
                    if (!Signature.access$200(firstService)) {
                        continue;
                    }
                    try {
                        this.sigSpi = newInstance(firstService);
                        this.provider = firstService.getProvider();
                        this.firstService = null;
                        this.serviceIterator = null;
                        return;
                    }
                    catch (NoSuchAlgorithmException ex) {
                        cause = ex;
                    }
                }
                ProviderException o = new ProviderException("Could not construct SignatureSpi instance");
                if (cause != null) {
                    o.initCause(cause);
                }
                throw o;
            }
        }
        
        private void chooseProvider(int n, Key key, AlgorithmParameterSpec algorithmParameterSpec, SecureRandom secureRandom) throws InvalidKeyException, InvalidAlgorithmParameterException {
            synchronized (this.lock) {
                if (this.sigSpi != null) {
                    return;
                }
                Exception ex = null;
                while (this.firstService != null || this.serviceIterator.hasNext()) {
                    Provider.Service firstService;
                    if (this.firstService != null) {
                        firstService = this.firstService;
                        this.firstService = null;
                    }
                    else {
                        firstService = (Provider.Service)this.serviceIterator.next();
                    }
                    if (key != null && !firstService.supportsParameter(key)) {
                        continue;
                    }
                    if (!Signature.access$200(firstService)) {
                        continue;
                    }
                    try {
                        SignatureSpi instance = newInstance(firstService);
                        this.tryOperation(instance, n, key, algorithmParameterSpec, secureRandom);
                        this.provider = firstService.getProvider();
                        this.sigSpi = instance;
                        this.firstService = null;
                        this.serviceIterator = null;
                        return;
                    }
                    catch (Exception ex2) {
                        if (ex == null) {
                            ex = ex2;
                        }
                        else {
                            continue;
                        }
                    }
                }
                if ((ex instanceof InvalidKeyException)) {
                    throw (InvalidKeyException)ex;
                }
                if ((ex instanceof RuntimeException)) {
                    throw (RuntimeException)ex;
                }
                if ((ex instanceof InvalidAlgorithmParameterException)) {
                    throw (InvalidAlgorithmParameterException)ex;
                }
                throw new InvalidKeyException("No installed provider supports this key: " + ((key != null) ? key.getClass().getName() : "(null)"), ex);
            }
        }
        
        private void tryOperation(SignatureSpi signatureSpi, int i, Key key, AlgorithmParameterSpec algorithmParameterSpec, SecureRandom secureRandom) throws InvalidKeyException, InvalidAlgorithmParameterException {
            switch (i) {
                case 1: {
                    signatureSpi.engineInitVerify((PublicKey)key);
                    break;
                }
                case 4: {
                    signatureSpi.engineInitVerify((PublicKey)key, algorithmParameterSpec);
                    break;
                }
                case 2: {
                    signatureSpi.engineInitSign((PrivateKey)key);
                    break;
                }
                case 3: {
                    signatureSpi.engineInitSign((PrivateKey)key, secureRandom);
                    break;
                }
                case 5: {
                    signatureSpi.engineInitSign((PrivateKey)key, algorithmParameterSpec, secureRandom);
                    break;
                }
                case 6: {
                    signatureSpi.engineSetParameter(algorithmParameterSpec);
                    break;
                }
                default: {
                    throw new AssertionError("Internal error: " + i);
                }
            }
        }
        
        @Override
        protected void engineInitVerify(PublicKey publicKey) throws InvalidKeyException {
            if (this.sigSpi != null) {
                this.sigSpi.engineInitVerify(publicKey);
            }
            else {
                try {
                    this.chooseProvider(1, publicKey, null, null);
                }
                catch (InvalidAlgorithmParameterException ex) {
                    throw new InvalidKeyException(ex);
                }
            }
        }
        
        @Override
        void engineInitVerify(PublicKey publicKey, AlgorithmParameterSpec algorithmParameterSpec) throws InvalidKeyException, InvalidAlgorithmParameterException {
            if (this.sigSpi != null) {
                this.sigSpi.engineInitVerify(publicKey, algorithmParameterSpec);
            }
            else {
                this.chooseProvider(4, publicKey, algorithmParameterSpec, null);
            }
        }
        
        @Override
        protected void engineInitSign(PrivateKey privateKey) throws InvalidKeyException {
            if (this.sigSpi != null) {
                this.sigSpi.engineInitSign(privateKey);
            }
            else {
                try {
                    this.chooseProvider(2, privateKey, null, null);
                }
                catch (InvalidAlgorithmParameterException ex) {
                    throw new InvalidKeyException(ex);
                }
            }
        }
        
        @Override
        protected void engineInitSign(PrivateKey privateKey, SecureRandom secureRandom) throws InvalidKeyException {
            if (this.sigSpi != null) {
                this.sigSpi.engineInitSign(privateKey, secureRandom);
            }
            else {
                try {
                    this.chooseProvider(3, privateKey, null, secureRandom);
                }
                catch (InvalidAlgorithmParameterException ex) {
                    throw new InvalidKeyException(ex);
                }
            }
        }
        
        @Override
        void engineInitSign(PrivateKey privateKey, AlgorithmParameterSpec algorithmParameterSpec, SecureRandom secureRandom) throws InvalidKeyException, InvalidAlgorithmParameterException {
            if (this.sigSpi != null) {
                this.sigSpi.engineInitSign(privateKey, algorithmParameterSpec, secureRandom);
            }
            else {
                this.chooseProvider(5, privateKey, algorithmParameterSpec, secureRandom);
            }
        }
        
        @Override
        protected void engineUpdate(byte b) throws SignatureException {
            this.chooseFirstProvider();
            this.sigSpi.engineUpdate(b);
        }
        
        @Override
        protected void engineUpdate(byte[] array, int n, int n2) throws SignatureException {
            this.chooseFirstProvider();
            this.sigSpi.engineUpdate(array, n, n2);
        }
        
        @Override
        protected void engineUpdate(ByteBuffer byteBuffer) {
            this.chooseFirstProvider();
            this.sigSpi.engineUpdate(byteBuffer);
        }
        
        @Override
        protected byte[] engineSign() throws SignatureException {
            this.chooseFirstProvider();
            return this.sigSpi.engineSign();
        }
        
        @Override
        protected int engineSign(byte[] array, int n, int n2) throws SignatureException {
            this.chooseFirstProvider();
            return this.sigSpi.engineSign(array, n, n2);
        }
        
        @Override
        protected boolean engineVerify(byte[] array) throws SignatureException {
            this.chooseFirstProvider();
            return this.sigSpi.engineVerify(array);
        }
        
        @Override
        protected boolean engineVerify(byte[] array, int n, int n2) throws SignatureException {
            this.chooseFirstProvider();
            return this.sigSpi.engineVerify(array, n, n2);
        }
        
        @Override
        protected void engineSetParameter(String s, Object o) throws InvalidParameterException {
            this.chooseFirstProvider();
            this.sigSpi.engineSetParameter(s, o);
        }
        
        @Override
        protected void engineSetParameter(AlgorithmParameterSpec algorithmParameterSpec) throws InvalidAlgorithmParameterException {
            if (this.sigSpi != null) {
                this.sigSpi.engineSetParameter(algorithmParameterSpec);
            }
            else {
                try {
                    this.chooseProvider(6, null, algorithmParameterSpec, null);
                }
                catch (InvalidKeyException ex) {
                    throw new InvalidAlgorithmParameterException(ex);
                }
            }
        }
        
        @Override
        protected Object engineGetParameter(String s) throws InvalidParameterException {
            this.chooseFirstProvider();
            return this.sigSpi.engineGetParameter(s);
        }
        
        @Override
        protected AlgorithmParameters engineGetParameters() {
            this.chooseFirstProvider();
            return this.sigSpi.engineGetParameters();
        }
        
        static {
            Signature.Delegate.warnCount = 10;
        }
    }
    
    private static class CipherAdapter extends SignatureSpi
    {
        private final Cipher cipher;
        private ByteArrayOutputStream data;
        
        CipherAdapter(Cipher cipher) {
            this.cipher = cipher;
        }
        
        @Override
        protected void engineInitVerify(PublicKey key) throws InvalidKeyException {
            this.cipher.init(2, key);
            if (this.data == null) {
                this.data = new ByteArrayOutputStream(128);
            }
            else {
                this.data.reset();
            }
        }
        
        @Override
        protected void engineInitSign(PrivateKey key) throws InvalidKeyException {
            this.cipher.init(1, key);
            this.data = null;
        }
        
        @Override
        protected void engineInitSign(PrivateKey key, SecureRandom random) throws InvalidKeyException {
            this.cipher.init(1, key, random);
            this.data = null;
        }
        
        @Override
        protected void engineUpdate(byte b) throws SignatureException {
            this.engineUpdate(new byte[] { b }, 0, 1);
        }
        
        @Override
        protected void engineUpdate(byte[] array, int n, int n2) throws SignatureException {
            if (this.data != null) {
                this.data.write(array, n, n2);
                return;
            }
            byte[] update = this.cipher.update(array, n, n2);
            if (update != null && update.length != 0) {
                throw new SignatureException("Cipher unexpectedly returned data");
            }
            return;
        }
        
        @Override
        protected byte[] engineSign() throws SignatureException {
            try {
                return this.cipher.doFinal();
            }
            catch (IllegalBlockSizeException ex) {
                throw new SignatureException("doFinal() failed", ex);
            }
            catch (BadPaddingException ex2) {
                throw new SignatureException("doFinal() failed", ex2);
            }
        }
        
        @Override
        protected boolean engineVerify(byte[] input) throws SignatureException {
            try {
                byte[] doFinal = this.cipher.doFinal(input);
                byte[] byteArray = this.data.toByteArray();
                this.data.reset();
                return MessageDigest.isEqual(doFinal, byteArray);
            }
            catch (BadPaddingException ex) {
                return false;
            }
            catch (IllegalBlockSizeException ex2) {
                throw new SignatureException("doFinal() failed", ex2);
            }
        }
        
        @Override
        protected void engineSetParameter(String s, Object o) throws InvalidParameterException {
            throw new InvalidParameterException("Parameters not supported");
        }
        
        @Override
        protected Object engineGetParameter(String s) throws InvalidParameterException {
            throw new InvalidParameterException("Parameters not supported");
        }
    }
}
